/******************************************************************************\
*
*	Apple Macintosh Developer Technical Support
*
*	Source file for the GMonde application
*
*	Program:	GMonde
*	File:		GMonde.c - C Implementation
*
*	by:			Forrest Tanaka
*
*	Copyright  1988-1991 Apple Computer, Inc.
*	All rights reserved.
*
\******************************************************************************/

#ifdef applec
#include <Devices.h>
#include <Events.h>
#include <Fonts.h>
#include <Memory.h>
#include <Menus.h>
#include <OSUtils.h>
#include <QuickDraw.h>
#include <Script.h>
#include <ToolUtils.h>
#include <Windows.h>
#endif
#include <AppleEvents.h>
#include <GestaltEqu.h>
#include <QDOffscreen.h>
#include <Palettes.h>
#include <Video.h>


/******************************************************************************\
* Constant Declarations
\******************************************************************************/

/* Resource IDs */
#define rDemoWindID    128 /* Demonstration window WIND resource ID */
#define rMarqueePattID 128 /* Marquee pattern PAT  resource ID */

/* Specifications for drawing the color ramps */
#define kMaxRampCount 256 /* Number of steps in a ramp */
#define kRampWidth    16  /* Width of a ramp in pixels */

/* Miscellaneous constants */
#define kMinScrollTime 1    /* # ticks between marquee pattern scrolls */
#define kResumeMask    1    /* Mask used to tell if got suspend/resume event */
#define kUseTempMem    true /* Pass to CreateOffScreen; GWorld in temp mem */

#ifdef applec
#define thePort    qd.thePort
#define gray       qd.gray
#define screenBits qd.screenBits
#endif

/* Macros to get the top-left and bottom-right corners of a rectangle */
#ifdef applec
#define topLeft(r) (*((Point *) &(r).top))
#define botRight(r) (*((Point *) &(r).bottom))
#endif

 
/******************************************************************************\
* Global Variable Definitions
\******************************************************************************/

WindowPtr   gDemoWindow;      /* Pointer to display window */
GDHandle    gDemoDevice;      /* Handle to the normal GDevice */
GWorldPtr   gOffScreen;       /* Pointer to off-screen GWorld */
Boolean     gQuitting;        /* True if GMonde is quitting */
Boolean     gHasGWorlds;      /* True if GWorlds are implemented */
Boolean     gHasCQD;          /* True if Color QuickDraw is implemented */
CTabHandle  gColors;          /* Colors to use for palette and off-screen */


/******************************************************************************\
* Function Prototypes
\******************************************************************************/

void main(void);

void StartUp(void);

void MainEventLoop(void);

pascal void HandleAEquit(
	AppleEvent *quitAppleEvent,
	AppleEvent *reply,
	long       handlerRefCon);

OSErr DoneRequiredParams(
	AppleEvent *anAppleEvent);

WindowPtr CreateDemoWindow(void);

GWorldPtr CreateOffScreen(
	GrafPtr    baseWindow,
	short      depth,
	CTabHandle colors);

void PaintVImage(void);

void PaintHImage(void);

extern pascal OSErr VideoStatus (short refNum, short csCode, Ptr csParamPtr);


/******************************************************************************\
* Main Program
\******************************************************************************/

void main()
{
	StartUp();
	if (gHasGWorlds)
	{
		gDemoWindow = CreateDemoWindow();
		if (gDemoWindow != nil)
		{
			SetPort( gDemoWindow );

			gOffScreen = CreateOffScreen( gDemoWindow, 8, nil );
			if (gOffScreen != nil)
			{
				GetGWorld( /*<*/(CGrafPtr *) &gDemoWindow, /*<*/&gDemoDevice );
				SetGWorld( gOffScreen, (GDHandle)nil );
				if (LockPixels( GetGWorldPixMap( gOffScreen ) ))
				{
					PaintVImage();
					UnlockPixels( GetGWorldPixMap( gOffScreen ) );
				}
				SetGWorld( (CGrafPtr)gDemoWindow, gDemoDevice );
				MainEventLoop();
			}
		}
	}
}


/******************************************************************************\
* StartUp - Start up our application
*
*     The usual things that have to be done as soon as an application starts are
* done here.  A few of the managers are initialized, and then the graphics
* environment is checked.
*
*     The global variable gHasCQD is set to true if the machine GMonde is
* running on has Color QuickDraw, false if not.  This is determined through
* Gestalt.  If Gestalt isnt working for some reason, then SysEnvirons is used
* instead.  The global variable gHasGWorlds is set to true if the GWorlds are
* implemented on the machine that GMonde is running on.  GWorlds are implemented
* if Gestalt returns gestalt32BitQD or greater, and if Gestalt returns a value
* between gestaltOriginalQD and gestalt8BitQD, exclusive.  Gestalt returns a
* value in this range if GMonde is running on a classic QuickDraw machine
* running system software version 7.0 or greater.
*
*     The AppleEvent quit handler is installed here.  When an incoming quit
* AppleEvent is processed, control is routed to the HandleAEquit function thats
* defined elsewhere in this program.
\******************************************************************************/

static void StartUp()
{
	long          qdVersion; /* Version of QuickDraw on this machine */
	long          aeAttr;    /* AppleEvents attributes */
	OSErr         result;

	InitGraf( &thePort );
	InitFonts();
	InitWindows();
	InitMenus();

	/* Find out if GWorlds and CQD are implemented on this machine */
	(void)Gestalt( gestaltQuickdrawVersion, /*<*/&qdVersion );
	gHasGWorlds = (qdVersion > gestaltOriginalQD && qdVersion < gestalt8BitQD)
	              || qdVersion >= gestalt32BitQD;
	gHasCQD = qdVersion >= gestalt8BitQD;

	/* Install the AppleEvent handler */
	result = Gestalt( gestaltAppleEventsAttr, /*<*/&aeAttr );
	if (result == noErr && (aeAttr & (1L << gestaltAppleEventsPresent)))
		result = AEInstallEventHandler( kCoreEventClass, kAEQuitApplication,
				(EventHandlerProcPtr)HandleAEquit, 0, false );
}


/******************************************************************************\
* MainEventLoop - Main event loop
*
*     This function collects events, does something on events that it knows
* about, and exits when the user clicks on the close box of the window.
*
*     If the user clicks in the content region of the window, then TrackMarquee
* is called to allow the user to drag out a selection marquee.
*
*     If the user clicks in the title bar, the user is allowed to drag the
* window around in the usual manner.  After the window is moved, UpdateOffScreen
* is called so that if the window was moved to a screen with a depth different
* than the one it was on, then the off-screen GWorld containing the image of the
* ramps is updated to the depth of that screen.  If the window is moved so that
* it crosses more than one screen, then the off-screen GWorld is set to the
* depth and color table of the deepest screen thats crossed.  That leads to an
* interesting situation if the window is moved so that its partially on a
* screen of a deeper depth than the one it was entirely contained on because the
* depth of the GWorld changes, yet the part of the image that remains on the old
* screen is from the old depth of the GWorld.  That can lead to some weird
* window updating problems.  So in that case, I invalidate the entire window so
* that all of the image is redrawn from the new GWorld depth.
*
*     If an update event comes in for the display window, UpdateOffScreen is
* called in case the update event is caused by the user using the Monitors
* control panel to change the depth or color table of the screen that the window
* is on.  Similar to the case in which the window is dragged so that its partly
* on a deeper screen, the entire window is invalidated if the GWorlds depth or
* color table changes.
*
*     If a suspend event comes in, then the GWorld is copied to the window so
* that the selection marquee, if any, is erased.  When GMonde is resumed, the
* usual marching ants animation resumes.
*
*     If theres an active selection marquee, then the marquee pattern is
* rotated one pixel and the marquee rectangle is redrawn using this rotated
* pattern.  This is done once per iteration of the main event loop.  To make the
* ants march as smoothly as possible, the WaitNextEvent sleep time is set to 0
* to give GMonde as much time as possible.  If there is no active selection,
* then its kind of impolite to ask for that much time, so the sleep time is set
* to 60 if theres no selection.
\******************************************************************************/

static void MainEventLoop()
{
	EventRecord anEvent;       /* Incoming events */
	WindowPtr   evtWind;       /* Pointer to window relevant to event */
	short       clickArea;     /* Area that mouse was clicked (i.e. menu bar) */
	Byte        osEvtKind;     /* Kind of OSEvent; mouse-moved or suspend/resume */
	Rect        screenRect;    /* Rectangle of screen in global coordinates */
	long        scrollTime;    /* TickCount @ time of last marquee patt scroll */
	Boolean     wereInFront;   /* True if GMonde is front-most application */
	QDErr       error;         /* Result of QuickDraw operations */
	Boolean     firstPictDisp; /* True if displaying first image */
	VDGammaRecord    currentGamma;
	GammaTblPtr      currentGammaTable;
	GammaTblPtr      newGammaTable;
	GammaTblPtr      savedGammaTable;
	short            channelIndex;
	unsigned char    *newTable;
	unsigned char    *savedTable;
	short            count;
	long             fadeAmount;
	VDGammaRecord    newGamma;
	VDSetEntryRecord setEntriesRec;
	Ptr				 csPtr;
	GDHandle         mainDevice;
	CTabHandle       mainDeviceColors;

	/* Initialize our variables for the main event loop */
	gQuitting = false;
	error = noErr;
	scrollTime = 0;
	wereInFront = true;
	firstPictDisp = false;

	/* Were going to modify the gamma of the main device, so get it */
	mainDevice = GetMainDevice ();

	/* Continue the main event loop until user quits or an error occurs */
	InitCursor();
	while ((!gQuitting) && (error == noErr))
	{
		/* If a non-null event, handle it; otherwise, just march the ants */
		if (WaitNextEvent( everyEvent, /*<*/&anEvent, 60,
				(RgnHandle)nil ))
		{
			if (anEvent.what == mouseDown)
			{
				/* React to content clicks, title clicks, close box clicks */
				clickArea = FindWindow( anEvent.where, /*<*/&evtWind );
				if (clickArea == inContent)
				{
					SetPort( evtWind );

					/* Grab the main devices gamma table; VideoStatus in Sample.a */
					error = VideoStatus( (**mainDevice).gdRefNum, cscGetGamma,
							(Ptr)&currentGamma );

					/* Grab a pointer to the main devices gamma table */
					currentGammaTable = (GammaTblPtr)currentGamma.csGTable;

					/* Alloc mem for new gamma table */
					savedGammaTable = (GammaTblPtr)NewPtr( sizeof (GammaTbl) +
							currentGammaTable->gDataCnt * currentGammaTable->gChanCnt +
							currentGammaTable->gFormulaSize );

					/* Copy main devices gamma into new gamma except for table itself */
					BlockMove( (Ptr)currentGammaTable, (Ptr)savedGammaTable,
							sizeof (GammaTbl) + currentGammaTable->gDataCnt *
							currentGammaTable->gChanCnt +
							currentGammaTable->gFormulaSize );

					/* Alloc mem for new gamma table */
					newGammaTable = (GammaTblPtr)NewPtr( sizeof (GammaTbl) +
							currentGammaTable->gDataCnt * currentGammaTable->gChanCnt +
							currentGammaTable->gFormulaSize );

					/* Copy main devices gamma into new gamma except for table itself */
					BlockMove( (Ptr)savedGammaTable, (Ptr)newGammaTable,
							sizeof (GammaTbl) );

					/* Fade out the screen */
					for (fadeAmount = 100; fadeAmount >= 0; fadeAmount -= 1)
					{
						/* Get pointers to the actual tables */
						newTable = (unsigned char *)&newGammaTable->gFormulaData +
								newGammaTable->gFormulaSize;
						savedTable = (unsigned char *)&savedGammaTable->gFormulaData +
								savedGammaTable->gFormulaSize;

						/* For each channel in the table, fade it */
						for (channelIndex = 0; channelIndex < savedGammaTable->gChanCnt; ++channelIndex)
						{
							/* For each entry in each table channel, fade it */
							for (count = 0; count < savedGammaTable->gDataCnt; count++)
								newTable[count] = savedTable[count] * fadeAmount / 100L;

							/* Point to the next channels */
							newTable += newGammaTable->gDataCnt;
							savedTable += savedGammaTable->gDataCnt;
						}

						/* Write the new gamma table to the main device */
						newGamma.csGTable = (Ptr)newGammaTable;
						csPtr = (Ptr) &newGamma;
						error = Control( (**mainDevice).gdRefNum, cscSetGamma, (Ptr)&csPtr );

						/* Force the main device to use the new gamma table */
						mainDeviceColors = (**(**mainDevice).gdPMap).pmTable;
						setEntriesRec.csTable = (ColorSpec *)&(**mainDeviceColors).ctTable;
						setEntriesRec.csStart = 0;
						setEntriesRec.csCount = (**mainDeviceColors).ctSize;
						csPtr = (Ptr) &setEntriesRec;
						error = Control ((**mainDevice).gdRefNum, cscSetEntries,
								(Ptr) &csPtr);
					}

					/* Paint the appropriate image */
					if (firstPictDisp)
						PaintVImage();
					else
						PaintHImage();

					/* Fade in the screen */
					for (fadeAmount = 0; fadeAmount < 100; fadeAmount += 1)
					{
						/* Get pointers to the actual tables */
						newTable = (unsigned char *)&newGammaTable->gFormulaData +
								newGammaTable->gFormulaSize;
						savedTable = (unsigned char *)&savedGammaTable->gFormulaData +
								savedGammaTable->gFormulaSize;

						/* For each channel in the table, fade it */
						for (channelIndex = 0; channelIndex < savedGammaTable->gChanCnt; ++channelIndex)
						{
							/* For each entry in each table channel, fade it */
							for (count = 0; count < savedGammaTable->gDataCnt; count++)
								newTable[count] = savedTable[count] * fadeAmount / 100L;

							/* Point to the next channels */
							newTable += newGammaTable->gDataCnt;
							savedTable += savedGammaTable->gDataCnt;
						}

						/* Write the new gamma table to the main device */
						newGamma.csGTable = (Ptr) newGammaTable;
						csPtr = (Ptr) &newGamma;
						error = Control ((**mainDevice).gdRefNum, cscSetGamma, (Ptr) &csPtr);

						/* Force the main device to use the new gamma table */
						mainDeviceColors = (**(**mainDevice).gdPMap).pmTable;
						setEntriesRec.csTable = (ColorSpec *)&(**mainDeviceColors).ctTable;
						setEntriesRec.csStart = 0;
						setEntriesRec.csCount = (**mainDeviceColors).ctSize;
						csPtr = (Ptr) &setEntriesRec;
						error = Control ((**mainDevice).gdRefNum, cscSetEntries,
								(Ptr) &csPtr);
					}

					DisposPtr( (Ptr)newGammaTable );
					DisposPtr( (Ptr)savedGammaTable );

					/* Choose the other image the next time round */
					firstPictDisp = !firstPictDisp;
				}

				else if (clickArea == inDrag)
				{
					/* Let the user drag the window around */
					screenRect = (**GetGrayRgn()).rgnBBox;
					DragWindow( evtWind, anEvent.where, &screenRect );
				}

				else if (clickArea == inGoAway)
					/* Click in close box means were outta here */
					if (TrackGoAway( evtWind, anEvent.where ))
						gQuitting = true;
			}
			else if (anEvent.what == updateEvt)
			{
				evtWind = (WindowPtr) anEvent.message;

				BeginUpdate( evtWind );
				SetPort( evtWind );

				/* Now draw the GWorld to the window */
				if (LockPixels( GetGWorldPixMap( gOffScreen ) ))
				{
					CopyBits( &((GrafPtr)gOffScreen)->portBits,
							&evtWind->portBits,
							&gOffScreen->portRect,
							&evtWind->portRect,
							srcCopy, nil );
					UnlockPixels( GetGWorldPixMap( gOffScreen ) );
				}

				EndUpdate( evtWind );
			}

			else if (anEvent.what == osEvt)
			{
				osEvtKind = (anEvent.message >> 24) & 0x00FF;
				if ((osEvtKind == suspendResumeMessage) &&
						(anEvent.message & kResumeMask))
					wereInFront = true;
				else
					wereInFront = false;
			}

			else if (anEvent.what == kHighLevelEvent)
				(void)AEProcessAppleEvent( &anEvent );
		}
	}
}


/******************************************************************************\
* HandleAEquit - Handler for 'quit' AppleEvent
*
*     This is the AppleEvent handler for the 'quit' AppleEvent as passed in the
* quitAppleEvent parameter by the AppleEvent Manager.  The DoQuit routine is
* called which causes this application to quit at the start of the next
* iteration of the main event loop.
*
*     Though the quit AppleEvent doesnt contain any parameters, the standard
* thing to do in reaction to any AppleEvent is to check to see if there are any
* required parameters in the AppleEvent that this routine doesnt recognise.
* DoneRequiredParms checks for this condition and returns an error if there are
* in fact required parameters in the AppleEvent or if some other error occurs
* during the check.
\******************************************************************************/

static pascal void HandleAEquit(
	AppleEvent *quitAppleEvent, /* Contains the quit AppleEvent */
	AppleEvent *reply,          /* Returns reply; ignored */
	long       handlerRefCon)   /* Application-defined parameter; ignored */
{
#pragma unused (reply, handlerRefCon)
	/* quit AE has no parms, but check in case the client requires any */
	if (DoneRequiredParams( quitAppleEvent ) == noErr)
		gQuitting = true;
}


/******************************************************************************\
* DoneRequiredParams - Done processing required params; OK?
*
*     DoneRequiredParams checks to see if the AppleEvent specified by the
* anAppleEvent parameter has any required parameters that we havent yet
* processed.  If there arent any left, then noErr is returned.  If there are
* required parameters that havent been processed yet, then errAEEventNotHandled
* is returned.  If any other errors occur, then that error code is returned.
\******************************************************************************/

static OSErr DoneRequiredParams(
	AppleEvent *anAppleEvent) /* AppleEvent being checked */
{
	DescType typeCode;   /* Type of AppleEvent attribute found; ignored */
	Size     actualSize; /* Actual size of parameters; ignored */
	OSErr    error;

	/* Are there any required parameters in AppleEvent we didnt process? */
   	error = AEGetAttributePtr( anAppleEvent, keyMissedKeywordAttr, typeWildCard,
			/*<*/&typeCode, nil, 0, /*<*/&actualSize );
	if (error == errAEDescNotFound)
		/* No required parameters left, so no error */
		error = noErr;
	else if (error == noErr)
		/* There was at least one required parameter we didnt process */
		error = errAEEventNotHandled;
	return error;
}


/******************************************************************************\
* CreateDemoWindow - Create a centered demonstration window
*
*     A new window is created and displayed on the main screen.  It is
* horizontally centered and on the upper third of the main screen.  A pointer to
* this window is returned.  This window is also set as the current GrafPort
* before it returns.  If the window couldnt be created for some reason, nil is
* returned.
*
*     On Color QuickDraw machines, a color window is created.  On machines
* without Color QuickDraw, a normal window is created.  The gHasCQD global
* variable is used to determine whether Color QuickDraw is implemented on this
* machine or not, so gHasCQD must be properly initialized before
* CreateDemoWindow is called.
\******************************************************************************/

static WindowPtr CreateDemoWindow ()
{
	WindowPtr aWindow;   /* Pointer to new Marching Ants window */
	Ptr       windStore; /* Pointer to WindowRecord memory */

	/* Assume failure, as I always do in life */
	aWindow = nil;

	/* Create window storage before GetNewWindow to avoid heap fragmentation */
	windStore = NewPtr( sizeof (WindowRecord) );
	if (windStore != nil)
		{
		/* If CQD, make color window, otherwise make normal window */
		if (gHasCQD)
			aWindow = GetNewCWindow( rDemoWindID, windStore, (WindowPtr)-1 );
		else
			aWindow = GetNewWindow( rDemoWindID, windStore, (WindowPtr)-1 );

		if (aWindow != nil)
		{
			/* Center window on the main screen */
			MoveWindow( aWindow,
					((screenBits.bounds.right - screenBits.bounds.left) -
					(aWindow->portRect.right - aWindow->portRect.left)) / 2,
					((screenBits.bounds.bottom - screenBits.bounds.top -
					GetMBarHeight()) - (aWindow->portRect.bottom -
					aWindow->portRect.top)) / 3 + GetMBarHeight(),
					false );

			/* Make the window visible and make it the current GrafPort */
			ShowWindow( aWindow );
			SetPort( aWindow );
		}
		else
			DisposPtr( windStore );
	}
	return aWindow;
}


/******************************************************************************\
* CreateOffScreen - Create an off-screen GWorld to buffer a windows image
*
*     An off-screen GWorld is created so that it can be used to buffer the image
* in the GrafPort specified by basePort.  To do that, I take advantage of
* NewGWorlds pixel-alignment capabilities.  If you pass zero as the pixel depth
* in the depth parameter and convert the portRect of basePort to global
* coordinates, the GWorld is created a pixel depth equal to the deepest screen
* that the portRect intersects, a color table to match that of the screens, and
* pixels aligned so that a CopyBits call to copy pixels between the screen and
* this GWorld is as fast as possible.
*
*     Starting with 32-Bit QuickDraw 1.2 (which was first included with system
* software version 6.0.5), you could ask NewGWorld to allocate the pixel image
* in temporary memory (formerly called MultiFinder Temporary Memory) by
* passing useTempMem in the GWorld flags.  If you pass true in temporary,
* NewGWorld will be called with the useTempMem flag set.  Otherwise, NewGWorld
* will be called normally.  As usual, pixel images allocated in temporary memory
* should only be used for transient GWorlds.
*
*     I set the clip region of the new GWorld to be coincident with its portRect
* as a matter of habit as soon as I create a GWorld.  The reason for this is
* because UpdateGWorld, at least as of this writing, resizes the clip region of
* a GWorld if you specify the stretchBits flag.  NewGWorld initializes the clip
* region to the entire QuickDraw coordinate plane.  When UpdateGWorld resizes
* the GWorld larger with the stretchBits flag on, it also resizes the GWorlds
* clip region proportionately larger.  If the clip region already covers the
* entire coordinate plane, the coordinates will get wrapped around and the clip
* region will be set to an empty region, and so nothing can be drawn into that
* GWorld.  By setting the clip region to the portRect right from the start, this
* problem never happens.
\******************************************************************************/

static GWorldPtr CreateOffScreen(
	WindowPtr  basePort, /* Pointer to the window to base off-screen on */
	short      depth,    /* Desired depth of pixel map in bits per pixel */
	CTabHandle colors)   /* Colors to use in the GWorld */
{
	CGrafPtr    currPort;    /* Pointer to the saved port */
	GDHandle    currGDevice; /* Handle to the current GDevice */
	GWorldPtr   offScreen;   /* Pointer to our GWorld */
	Rect        offRect;     /* portRect of basePort in global coordinates */
	QDErr       result;      /* Error return from NewGWorld */

	GetGWorld( /*<*/&currPort, /*<*/&currGDevice );
	SetPort( basePort );

	/* Globalize portRect of basePort */
	offRect = basePort->portRect;
	if (depth == 0)
	{
		LocalToGlobal( /**/&topLeft( offRect ) );
		LocalToGlobal( /**/&botRight( offRect ) );
	}

	/* Create the new off-screen port and set it as the current port */
	result = NewGWorld( /*<*/&offScreen, depth, &offRect, colors, nil, 0 );
	if (result == noErr)
	{
		SetGWorld( offScreen, (GDHandle)nil );

		/* Good idea to set the clip region to the portRect of the new GWorld */
		ClipRect( &offScreen->portRect );

		/* Clear the new GWorld to white */
		if (LockPixels( GetGWorldPixMap( offScreen ) ))
		{
			EraseRect( &offScreen->portRect );
			UnlockPixels( GetGWorldPixMap( offScreen ) );
		}
	}
	else
		offScreen = nil;

	/* Reset current GrafPort and GDevice to what it was when we were called */
	SetGWorld( currPort, currGDevice );
	return offScreen;
}


/******************************************************************************\
* PaintVImage - Paint color ramps into the current GrafPort
*
*     PaintVImage paints a series of color ramps into the current GrafPort.  To
* showcase the error diffusion dithering that CopyBits can perform on Color
* QuickDraw machines, the image isnt drawn directly to the current GrafPort if
* Color QuickDraw is available.  Instead, a temporary 32-bit deep GWorld is
* created and the ramps are drawn into it.  When thats done, the 32-bit deep
* GWorld is copied to the current GrafPort with the ditherMode modifier on.  If
* there isnt enough memory for the temporary GWorld or if Color QuickDraw isnt
* available, then the ramps are drawn directly to the current GrafPort.
\******************************************************************************/

static void PaintVImage ()
	{
	CGrafPtr savedPort;    /* Pointer to saved CGrafPort for later restoring */
	GDHandle savedGDevice; /* Handle to saved GDevice for later restoring */
	RGBColor rampColor;    /* Color to use when drawing a step of the ramp */
	short    rampFactor;   /* Value to set a color component to in a ramp */
	short    pinFactor;    /* Value to pin a color component to in a ramp */
	short    x;            /* Horizontal coordinate to draw ramp step */
	short    y;            /* Vertical coordinate to draw ramp step */

	GetGWorld( /*<*/&savedPort, /*<*/&savedGDevice );

	/* Pin component values at $0000 first, then at $FFFF */
	for (pinFactor = 0; pinFactor >= -1; pinFactor--)
	{
		/* Loop on y coordinate over entire GrafPort */
		for (y = 0; y < kMaxRampCount; y++)
		{
			/* Calculate value of components that are dynamic in a ramp */
			rampFactor = y * 65535 / kMaxRampCount;
			if (pinFactor == -1)
			{
				/* If pinning at $FFFF, then use ones comp of rampFactor */
				rampFactor ^= 0xFFFF;
				x = kRampWidth;
			}
			else
				x = 0;

			rampColor.red = rampFactor;
			rampColor.green = rampFactor;
			rampColor.blue = rampFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.blue = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.green = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.blue = rampFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.red = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.green = rampFactor;
			RGBForeColor( &rampColor );
			MoveTo(x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.blue = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );

			x += kRampWidth * 2;
			rampColor.green = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( x, y );
			Line( kRampWidth, 0 );
		}
	}
}


/******************************************************************************\
* PaintVImage - Paint color ramps into the current GrafPort
*
*     PaintVImage paints a series of color ramps into the current GrafPort.  To
* showcase the error diffusion dithering that CopyBits can perform on Color
* QuickDraw machines, the image isnt drawn directly to the current GrafPort if
* Color QuickDraw is available.  Instead, a temporary 32-bit deep GWorld is
* created and the ramps are drawn into it.  When thats done, the 32-bit deep
* GWorld is copied to the current GrafPort with the ditherMode modifier on.  If
* there isnt enough memory for the temporary GWorld or if Color QuickDraw isnt
* available, then the ramps are drawn directly to the current GrafPort.
\******************************************************************************/

static void PaintHImage ()
{
	CGrafPtr savedPort;    /* Pointer to saved CGrafPort for later restoring */
	GDHandle savedGDevice; /* Handle to saved GDevice for later restoring */
	RGBColor rampColor;    /* Color to use when drawing a step of the ramp */
	short    rampFactor;   /* Value to set a color component to in a ramp */
	short    pinFactor;    /* Value to pin a color component to in a ramp */
	short    x;            /* Horizontal coordinate to draw ramp step */
	short    y;            /* Vertical coordinate to draw ramp step */

	GetGWorld( /*<*/&savedPort, /*<*/&savedGDevice );

	/* Pin component values at $0000 first, then at $FFFF */
	for (pinFactor = 0; pinFactor >= -1; pinFactor--)
	{
		/* Loop on y coordinate over entire GrafPort */
		for (y = 0; y < kMaxRampCount; y++)
		{
			/* Calculate value of components that are dynamic in a ramp */
			rampFactor = y * 65535 / kMaxRampCount;
			if (pinFactor == -1)
			{
				/* If pinning at $FFFF, then use ones comp of rampFactor */
				rampFactor ^= 0xFFFF;
				x = kRampWidth;
			}
			else
				x = 0;

			rampColor.red = rampFactor;
			rampColor.green = rampFactor;
			rampColor.blue = rampFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.blue = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.green = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.blue = rampFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.red = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.green = rampFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.blue = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );

			x += kRampWidth * 2;
			rampColor.green = pinFactor;
			RGBForeColor( &rampColor );
			MoveTo( y, x );
			Line( 0, kRampWidth );
		}
	}
}
